
CutterMultiFruit = {}
CutterMultiFruit.modDir = g_currentModDirectory
CutterMultiFruit.modName = g_currentModName

local modDesc = loadXMLFile("modDesc", CutterMultiFruit.modDir .. "modDesc.xml")

CutterMultiFruit.usedXmlString = Utils.getNoNil(getXMLString(modDesc, "modDesc.LuaVariable#name"), "")



function CutterMultiFruit.initSpecialization()

	g_workAreaTypeManager:addWorkAreaType("cutter", false)

end



function CutterMultiFruit.prerequisitesPresent(specializations)

	return SpecializationUtil.hasSpecialization(WorkArea, specializations)

end



function CutterMultiFruit.registerFunctions(vehicleType)

	SpecializationUtil.registerFunction(vehicleType, "readTestAreasStream",					CutterMultiFruit.readTestAreasStream)
	SpecializationUtil.registerFunction(vehicleType, "writeTestAreasStream",				CutterMultiFruit.writeTestAreasStream)
	SpecializationUtil.registerFunction(vehicleType, "getCombine",							CutterMultiFruit.getCombine)
	SpecializationUtil.registerFunction(vehicleType, "getAllowCutterAIFruitRequirements",	CutterMultiFruit.getAllowCutterAIFruitRequirements)
	SpecializationUtil.registerFunction(vehicleType, "loadTestAreaFromXML",					CutterMultiFruit.loadTestAreaFromXML)
	SpecializationUtil.registerFunction(vehicleType, "getIsTestAreaActive",					CutterMultiFruit.getIsTestAreaActive)
	SpecializationUtil.registerFunction(vehicleType, "processCutterArea",					CutterMultiFruit.processCutterArea)
	SpecializationUtil.registerFunction(vehicleType, "processPickupCutterArea",				CutterMultiFruit.processPickupCutterArea)
	SpecializationUtil.registerFunction(vehicleType, "getCutterLoad",						CutterMultiFruit.getCutterLoad)

end



function CutterMultiFruit.registerOverwrittenFunctions(vehicleType)

	SpecializationUtil.registerOverwrittenFunction(vehicleType, "loadSpeedRotatingPartFromXML",		CutterMultiFruit.loadSpeedRotatingPartFromXML)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getIsSpeedRotatingPartActive",		CutterMultiFruit.getIsSpeedRotatingPartActive)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "loadRandomlyMovingPartFromXML",	CutterMultiFruit.loadRandomlyMovingPartFromXML)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getIsRandomlyMovingPartActive",	CutterMultiFruit.getIsRandomlyMovingPartActive)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getIsWorkAreaActive",				CutterMultiFruit.getIsWorkAreaActive)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "doCheckSpeedLimit",				CutterMultiFruit.doCheckSpeedLimit)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "loadWorkAreaFromXML",				CutterMultiFruit.loadWorkAreaFromXML)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getDirtMultiplier",				CutterMultiFruit.getDirtMultiplier)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getWearMultiplier",				CutterMultiFruit.getWearMultiplier)

end



function CutterMultiFruit.registerEventListeners(vehicleType)

	SpecializationUtil.registerEventListener(vehicleType, "onLoad", CutterMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onPostLoad", CutterMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onDelete", CutterMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onReadStream", CutterMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onWriteStream", CutterMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onReadUpdateStream", CutterMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onWriteUpdateStream", CutterMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onUpdateTick", CutterMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onStartWorkAreaProcessing", CutterMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onEndWorkAreaProcessing", CutterMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onPreAttach", CutterMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onPostDetach", CutterMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onTurnedOn", CutterMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onTurnedOff", CutterMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onAIImplementStart", CutterMultiFruit)

end



function CutterMultiFruit:onLoad(savegame)

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.turnedOnRotationNodes.turnedOnRotationNode#type", "vehicle.cutter.animationNodes.animationNode", "cutter") --FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.turnedOnScrollers", "vehicle.cutter.animationNodes.animationNode") --FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.cutter.turnedOnScrollers", "vehicle.cutter.animationNodes.animationNode") --FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.cutter.reelspikes", "vehicle.cutter.rotationNodes.rotationNode or vehicle.turnOnVehicle.turnedOnAnimation") --FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.cutter.threshingParticleSystems.threshingParticleSystem", "vehicle.cutter.fillEffect.effectNode") --FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.cutter.threshingParticleSystems.emitterShape", "vehicle.cutter.fillEffect.effectNode") --FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.cutter#convertedFillTypeCategories", "vehicle.cutter#fruitTypeConverter") --FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.cutter#startAnimationName", "vehicle.turnOnVehicle.turnOnAnimation#name") --FS17 to FS19



	-- load fruitTypes
	local fruitTypes = nil
	local fruitTypeNames = getXMLString(self.xmlFile, "vehicle.cutter#fruitTypes")
	local fruitTypeCategories = getXMLString(self.xmlFile, "vehicle.cutter#fruitTypeCategories")

	if fruitTypeNames == CutterMultiFruit.usedXmlString then

		fruitTypeNames = ""

		for fruitTypeIndex, fruitTypeData in pairs(g_fruitTypeManager.fruitTypes) do

			if fruitTypeData ~= nil then

				local minHarvestingGrowthState = fruitTypeData.minHarvestingGrowthState
				local minPreparingGrowthState = fruitTypeData.minPreparingGrowthState
				local preparedGrowthState = fruitTypeData.preparedGrowthState
				local name = fruitTypeData.name

				if minHarvestingGrowthState == 9 and minPreparingGrowthState ~= -1 and preparedGrowthState == 9 and name ~= "SUGARCANE" then

					if fruitTypeNames ~= "" then 
						fruitTypeNames = fruitTypeNames .." "
					end

					fruitTypeNames = fruitTypeNames..tostring(name)

				end

			end

		end

	end

	if fruitTypeCategories ~= nil and fruitTypeNames == nil then

		fruitTypes = g_fruitTypeManager:getFruitTypesByCategoryNames(fruitTypeCategories, "Warning: Cutter has invalid fruitTypeCategory '%s' in '"..self.configFileName.."'")

	elseif fruitTypeCategories == nil and fruitTypeNames ~= nil then

		fruitTypes = g_fruitTypeManager:getFruitTypesByNames(fruitTypeNames, "Warning: Cutter has invalid fruitType '%s' in '"..self.configFileName.."'")

	else

		g_logManager:xmlWarning(self.configFileName, "Cutter needs either the 'fruitTypeCategories' or 'fruitTypes' attribute!")

	end

	if fruitTypes ~= nil then

		spec.fruitTypes = {}

		for _,fruitType in pairs(fruitTypes) do

			spec.fruitTypes[fruitType] = true

		end

	end



	spec.fruitTypeConverters = {}

	local category = getXMLString(self.xmlFile, "vehicle.cutter#fruitTypeConverter")

	if category ~= nil then

		local data = g_fruitTypeManager:getConverterDataByName(category)

		if data ~= nil then

			for input, converter in pairs(data) do

				spec.fruitTypeConverters[input] = converter

			end

		end

	end

	if self.isClient then

		spec.animationNodes = g_animationManager:loadAnimations(self.xmlFile, "vehicle.cutter.animationNodes", self.components, self, self.i3dMappings)
		spec.fruitExtraObjects = {}

		local i = 0

		while true do

			local key = string.format("vehicle.cutter.fruitExtraObjects.fruitExtraObject(%d)", i)

			XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, key.."#index", key.."#node")

			local node = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, key.."#node"), self.i3dMappings)
			local anim = getXMLString(self.xmlFile, key.."#anim")
			local isDefault = Utils.getNoNil(getXMLBool(self.xmlFile, key.."#isDefault"), false)
			local fruitType = g_fruitTypeManager:getFruitTypeByName(getXMLString(self.xmlFile, key.."#fruitType"))

			if fruitType == nil or (node == nil and anim == nil) then
				break
			end

			if node ~= nil then
				setVisibility(node, false)
			end

			local extraObject = {node=node, anim=anim}

			spec.fruitExtraObjects[fruitType.index] = extraObject

			if isDefault then
				spec.fruitExtraObjects[FruitType.UNKNOWN] = extraObject
			end

			i = i + 1

		end

		spec.hideExtraObjectsOnDetach = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.cutter.fruitExtraObjects#hideOnDetach"), false)
		spec.testAreas = {}
		spec.testAreaBase = Utils.getNoNil(I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.cutter.testAreas#baseNode"), self.i3dMappings), self.rootNode)

		i = 0

		while true do

			local key = string.format("vehicle.cutter.testAreas.testArea(%d)", i)

			if not hasXMLProperty(self.xmlFile, key) then
				break
			end

			local testArea = {}

			if self:loadTestAreaFromXML(testArea, self.xmlFile, key) then
				table.insert(spec.testAreas, testArea)
			end

			i = i + 1

		end



		-- cutter
		spec.cutterEffects = {}
		spec.currentCutterEffect = nil

		if hasXMLProperty(self.xmlFile, "vehicle.cutter.effect") then

			for fruitTypeIndex,_ in pairs(spec.fruitTypes) do

				spec.cutterEffects[fruitTypeIndex] = {}

				i = 0

				while true do

					local key = string.format("vehicle.cutter.effect.effectNode(%d)", i)

					if not hasXMLProperty(self.xmlFile, key) then
						break
					end

					local effect = CutterEffect:new()

					effect:load(self.xmlFile, key, self.components, self, fruitTypeIndex, self.i3dMappings)

					table.insert(spec.cutterEffects[fruitTypeIndex], effect)

					i = i + 1

				end

			end

		end

		spec.fillEffects = g_effectManager:loadEffect(self.xmlFile, "vehicle.cutter.fillEffect", self.components, self, self.i3dMappings)

	end



	spec.allowsForageGrowthState = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.cutter#allowsForageGrowthState"), false)
	spec.allowCuttingWhileRaised = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.cutter#allowCuttingWhileRaised"), false)
	spec.movingDirection = MathUtil.sign(Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.cutter#movingDirection"), 1))
	spec.strawRatio = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.cutter#strawRatio"), 1)
	spec.useWindrow = false
	spec.currentInputFillType = FillType.UNKNOWN
	spec.currentInputFruitType = FruitType.UNKNOWN
	spec.currentInputFruitTypeAI = FruitType.UNKNOWN
	spec.lastValidInputFruitType = FruitType.UNKNOWN
	spec.currentInputFruitTypeSent = FruitType.UNKNOWN
	spec.currentGrowthStateArea = 0
	spec.currentGrowthState = 0
	spec.lastAreaBiggerZero = false
	spec.lastAreaBiggerZeroSent = false
	spec.lastAreaBiggerZeroTime = -1
	spec.workAreaParameters = {}
	spec.workAreaParameters.lastRealArea = 0
	spec.workAreaParameters.lastArea = 0
	spec.workAreaParameters.lastGrowthState = 0
	spec.workAreaParameters.lastGrowthStateArea = 0
	spec.lastOutputFillTypes = {}
	spec.lastPrioritizedOutputType = FillType.UNKNOWN
	spec.lastOutputTime = 0
	spec.cutterLoad = 0
	spec.isWorking = false
	spec.workAreaParameters.countArea = true
	spec.aiNoValidGroundTimer = 0
	spec.dirtyFlag = self:getNextDirtyFlag()
	spec.effectDirtyFlag = self:getNextDirtyFlag()

	self:doCollisionMaskCheck(MathUtil.bitsToMask(26), "vehicle.cutter#ignoreCollisionMask")

end



function CutterMultiFruit:onPostLoad(savegame)

	if self.addCutterToCombine ~= nil then

		self:addCutterToCombine(self)

	end

	if self.isClient then

		CutterMultiFruit.updateExtraObjects(self)

	end

end



function CutterMultiFruit:onDelete()

	if self.isClient then

		local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

		for _, effectAttribute in pairs(spec.cutterEffects) do

			g_effectManager:deleteEffects(effectAttribute.effect)

		end

		g_effectManager:deleteEffects(spec.fillEffects)
		g_animationManager:deleteAnimations(spec.animationNodes)

	end

end



function CutterMultiFruit:onReadStream(streamId, timestamp, connection)

	self:readTestAreasStream(streamId, connection)

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

	spec.lastAreaBiggerZero = streamReadBool(streamId)

	if spec.lastAreaBiggerZero then

		spec.lastAreaBiggerZeroTime = g_currentMission.time

	end

end



function CutterMultiFruit:onWriteStream(streamId, connection)

	self:writeTestAreasStream(streamId, connection)

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

	streamWriteBool(streamId, spec.lastAreaBiggerZeroSent)

end



function CutterMultiFruit:onReadUpdateStream(streamId, timestamp, connection)

	if connection:getIsServer() then

		local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

		if streamReadBool(streamId) then

			self:readTestAreasStream(streamId, connection)

		end

		spec.lastAreaBiggerZero = streamReadBool(streamId)

		if spec.lastAreaBiggerZero then

			spec.lastAreaBiggerZeroTime = g_currentMission.time

		end

	end

end



function CutterMultiFruit:onWriteUpdateStream(streamId, connection, dirtyMask)

	if not connection:getIsServer() then

		local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

		if streamWriteBool(streamId, bitAND(dirtyMask, spec.effectDirtyFlag) ~= 0) then

			self:writeTestAreasStream(streamId, connection)

		end

		streamWriteBool(streamId, spec.lastAreaBiggerZeroSent)

	end

end



function CutterMultiFruit:readTestAreasStream(streamId, connection)

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

	local readGrowthState = false

	for _, testArea in ipairs(spec.testAreas) do

		testArea.hasFruitContact = streamReadBool(streamId)

		if testArea.hasFruitContact then

			readGrowthState = true

		end

	end

	if readGrowthState then
		spec.currentGrowthState = streamReadUIntN(streamId, 4)
	end

	spec.currentInputFruitType = streamReadUIntN(streamId, 7)

	if streamReadBool(streamId) then
		spec.lastValidInputFruitType = spec.currentInputFruitType
	else
		spec.currentInputFruitType = FillType.UNKNOWN
	end

	if streamReadBool(streamId) then
		spec.currentInputFillType = g_fruitTypeManager:getWindrowFillTypeIndexByFruitTypeIndex(spec.currentInputFruitType)
	else
		spec.currentInputFillType = g_fruitTypeManager:getFillTypeIndexByFruitTypeIndex(spec.currentInputFruitType)
	end

end



function CutterMultiFruit:writeTestAreasStream(streamId, connection)

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

	local sentGrowthState = false

	for _, testArea in ipairs(spec.testAreas) do

		streamWriteBool(streamId, testArea.hasFruitContact)

		if testArea.hasFruitContact then

			sentGrowthState = true

		end

	end

	if sentGrowthState then
		streamWriteUIntN(streamId, spec.currentGrowthState, 4)
	end

	streamWriteUIntN(streamId, spec.currentInputFruitType, 7)
	streamWriteBool(streamId, spec.currentInputFruitType == spec.lastValidInputFruitType)
	streamWriteBool(streamId, spec.useWindrow)

end



function CutterMultiFruit:onUpdateTick(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

	local isTurnedOn = self:getIsTurnedOn()
	local isEffectActive = isTurnedOn and self.movingDirection == spec.movingDirection and self:getLastSpeed() > 0.5 and (spec.allowCuttingWhileRaised or self:getIsLowered(true))

	if isEffectActive then

		local minEffectValue = math.huge
		local maxEffectValue = -math.huge
		local chargedAreas = 0

		for _, area in ipairs(spec.testAreas) do

			local x,y,z = getWorldTranslation(area.start)
			local x1,y1,z1 = getWorldTranslation(area.width)
			local x2,_,z2 = getWorldTranslation(area.height)

			if self.isServer and spec.currentInputFruitType ~= nil then

				if self:getIsTestAreaActive(area) then

					local fruitValue, _, _, _ = FSDensityMapUtil.getFruitArea(spec.currentInputFruitType, x, z, x1, z1, x2, z2, nil, spec.allowsForageGrowthState)

					area.hasFruitContact = fruitValue > 0

					if area.hasFruitContactSent ~= area.hasFruitContact then

						self:raiseDirtyFlags(spec.effectDirtyFlag)
						area.hasFruitContactSent = area.hasFruitContact

					end

				end

			end

			if area.hasFruitContact then

				local lx1,_,_ = worldToLocal(spec.testAreaBase, x,y,z)
				local lx2,_,_ = worldToLocal(spec.testAreaBase, x1,y1,z1)

				minEffectValue = math.min(minEffectValue, lx1, lx2)
				maxEffectValue = math.max(maxEffectValue, lx1, lx2)
				chargedAreas = chargedAreas + 1

			end

		end

		if not spec.useWindrow and #spec.testAreas > 0 then

			local currentLoad = chargedAreas / #spec.testAreas
			spec.cutterLoad = spec.cutterLoad * 0.95 + currentLoad * 0.05

		end

		local reset = false

		if minEffectValue == math.huge and maxEffectValue == -math.huge then

			minEffectValue = 0
			maxEffectValue = 0
			reset = true

		end

		if spec.movingDirection > 0 then

			minEffectValue = minEffectValue * -1
			maxEffectValue = maxEffectValue * -1

			if maxEffectValue < minEffectValue then

				local t = minEffectValue

				minEffectValue = maxEffectValue
				maxEffectValue = t

			end

		end

		local inputFruitType = spec.currentInputFruitType

		if inputFruitType ~= spec.lastValidInputFruitType then
			-- if we pickup a different fruit type than we are able to proceed to the combine we won't display the effect
			inputFruitType = nil
		end

		if inputFruitType ~= nil then

			local newEffect = spec.cutterEffects[inputFruitType]

			if newEffect ~= nil then

				if spec.currentCutterEffect ~= newEffect then
					g_effectManager:resetEffects(spec.currentCutterEffect)
				end

				spec.currentCutterEffect = newEffect

				for _, effect in ipairs(spec.currentCutterEffect) do

					if effect.setGrowthState ~= nil then
						effect:setGrowthState(spec.currentGrowthState)
					end

					if effect.setMinMaxWidth ~= nil then
						effect:setMinMaxWidth(minEffectValue, maxEffectValue, reset)
					end

				end

			else

				if spec.currentCutterEffect ~= nil then
					g_effectManager:resetEffects(spec.currentCutterEffect)
				end

			end

			CutterMultiFruit.updateExtraObjects(self)

		end

		local isCollecting = spec.lastAreaBiggerZeroTime + 300 > g_currentMission.time
		local fillType = spec.currentInputFillType

		if spec.useWindrow then

			if isCollecting then
				spec.cutterLoad = spec.cutterLoad * 0.95 + 0.05
			else
				spec.cutterLoad = spec.cutterLoad * 0.9
			end

		end

		if fillType ~= FillType.UNKNOWN and isCollecting then

			g_effectManager:setFillType(spec.fillEffects, fillType)
			g_effectManager:setMinMaxWidth(spec.fillEffects, minEffectValue, maxEffectValue, reset)
			g_effectManager:startEffects(spec.fillEffects)

		else

			g_effectManager:stopEffects(spec.fillEffects)

		end

	else

		if spec.currentCutterEffect ~= nil then

			for _, effect in pairs(spec.currentCutterEffect) do

				if effect.setMinMaxWidth ~= nil then
					effect:setMinMaxWidth(0, 0, true)
				end

			end

		end

		g_effectManager:stopEffects(spec.fillEffects)
		spec.cutterLoad = spec.cutterLoad * 0.9

	end

	-- lastPrioritizedOutputType is always the fill type that was the most significant fill type of the last 500ms
	-- this fill type is transfered to the combine to avoid quick changes of the fill type if we harvester 2 different fruit types at the same time
	-- e.g. on field borders if can happen that a bit of grass if collected from the cutter
	spec.lastOutputTime = spec.lastOutputTime + dt

	if spec.lastOutputTime > 500 then

		spec.lastPrioritizedOutputType = FillType.UNKNOWN
		local max = 0

		for i, _ in pairs(spec.lastOutputFillTypes) do

			if spec.lastOutputFillTypes[i] > max then

				spec.lastPrioritizedOutputType = i
				max = spec.lastOutputFillTypes[i]

			end

			spec.lastOutputFillTypes[i] = 0

		end

		spec.lastOutputTime = 0

	end

end



function CutterMultiFruit:getCombine()

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]
	local outputFillType = g_fruitTypeManager:getFillTypeIndexByFruitTypeIndex(spec.currentInputFruitType)

	if spec.fruitTypeConverters[spec.currentInputFruitType] ~= nil then

		outputFillType = spec.fruitTypeConverters[spec.currentInputFruitType].fillTypeIndex

	end

	if self.verifyCombine ~= nil then

		return self:verifyCombine(spec.currentInputFruitType, outputFillType)

	else

		if self.getAttacherVehicle ~= nil then

			local attacherVehicle = self:getAttacherVehicle()

			if attacherVehicle ~= nil then

				if attacherVehicle.verifyCombine ~= nil then

					return attacherVehicle:verifyCombine(spec.currentInputFruitType, outputFillType)

				end

			end

		end

	end

	return nil

end



function CutterMultiFruit:getAllowCutterAIFruitRequirements()

	return true

end



function CutterMultiFruit:processCutterArea(workArea, dt)

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

	if spec.workAreaParameters.combineVehicle ~= nil then

		local xs,_,zs = getWorldTranslation(workArea.start)
		local xw,_,zw = getWorldTranslation(workArea.width)
		local xh,_,zh = getWorldTranslation(workArea.height)
		local lastRealArea = 0
		local lastThreshedArea = 0
		local lastArea = 0

		for fruitType, state in pairs(spec.workAreaParameters.fruitTypesToUse) do

			if state then

				local realArea, area, sprayFactor, plowFactor, limeFactor, weedFactor, growthState, growthStateArea, terrainDetailPixelsSum = FSDensityMapUtil.cutFruitArea(fruitType, xs,zs, xw,zw, xh,zh, true, true, spec.allowsForageGrowthState, g_currentMission.chopperGroundLayerType)

				if realArea >  0 then

					if self.isServer then

						if growthStateArea > spec.currentGrowthStateArea then

							spec.currentGrowthStateArea = growthStateArea
							spec.currentGrowthState = growthState

						end

						spec.currentInputFruitType = fruitType

						-- ai only works on terrain detail, so we do not allow the ai to require fruits that are out of a field

						if terrainDetailPixelsSum > 0 then
							spec.currentInputFruitTypeAI = fruitType
						end

						spec.currentInputFillType = g_fruitTypeManager:getFillTypeIndexByFruitTypeIndex(fruitType)
						spec.useWindrow = false

					end

					local multiplier = g_currentMission:getHarvestScaleMultiplier(fruitType, sprayFactor, plowFactor, limeFactor, weedFactor)

					lastRealArea = realArea * multiplier
					lastThreshedArea = realArea
					lastArea = area
					spec.workAreaParameters.lastFruitType = fruitType

					break

				end

			end

		end

		if lastArea > 0 then

			if workArea.chopperAreaIndex ~= nil then

				local workArea = self:getWorkAreaByIndex(workArea.chopperAreaIndex)

				if workArea ~= nil then

					local xs,_,zs = getWorldTranslation(workArea.start)
					local xw,_,zw = getWorldTranslation(workArea.width)
					local xh,_,zh = getWorldTranslation(workArea.height)

					FSDensityMapUtil.setGroundTypeLayerArea(xs, zs, xw, zw, xh, zh, g_currentMission.chopperGroundLayerType)

				else

					workArea.chopperAreaIndex = nil
					g_logManager:xmlWarning(self.configFileName, "Invalid chopperAreaIndex '%d' for workArea '%d'!", workArea.chopperAreaIndex, workArea.index)

				end

			end

			spec.isWorking = true

		end

		spec.workAreaParameters.lastRealArea = spec.workAreaParameters.lastRealArea + lastRealArea
		spec.workAreaParameters.lastThreshedArea = spec.workAreaParameters.lastThreshedArea + lastThreshedArea
		spec.workAreaParameters.lastStatsArea = spec.workAreaParameters.lastStatsArea + lastThreshedArea
		spec.workAreaParameters.lastArea	 = spec.workAreaParameters.lastArea + lastArea

	end

	return spec.workAreaParameters.lastRealArea, spec.workAreaParameters.lastArea

end



function CutterMultiFruit:processPickupCutterArea(workArea, dt)

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

	if spec.workAreaParameters.combineVehicle ~= nil then

		local lsx, lsy, lsz, lex, ley, lez, lineRadius = DensityMapHeightUtil.getLineByArea(workArea.start, workArea.width, workArea.height)

		for fruitType, state in pairs(spec.workAreaParameters.fruitTypesToUse) do

			if state then

				local fillType = g_fruitTypeManager:getWindrowFillTypeIndexByFruitTypeIndex(fruitType)

				if fillType ~= nil then

					local pickedUpLiters = -DensityMapHeightUtil.tipToGroundAroundLine(self, -math.huge, fillType, lsx, lsy, lsz, lex, ley, lez, lineRadius, nil, nil, false, nil)

					if self.isServer then

						if pickedUpLiters > 0 then

							local fruitDesc = g_fruitTypeManager:getFruitTypeByIndex(fruitType)
							local literPerSqm = fruitDesc.literPerSqm
							local lastCutterArea = pickedUpLiters / (g_currentMission:getFruitPixelsToSqm() * literPerSqm);

							spec.currentInputFruitType = fruitType
							spec.useWindrow = true
							spec.currentInputFillType = fillType
							spec.workAreaParameters.lastFruitType = fruitType
							spec.workAreaParameters.lastRealArea  = spec.workAreaParameters.lastRealArea + lastCutterArea
							spec.workAreaParameters.lastThreshedArea  = spec.workAreaParameters.lastThreshedArea + lastCutterArea
							spec.workAreaParameters.lastStatsArea  = spec.workAreaParameters.lastStatsArea + lastCutterArea
							spec.workAreaParameters.lastArea	  = spec.workAreaParameters.lastArea + lastCutterArea
							spec.isWorking = true

							break

						end

					end

				end

			end

		end

	end

	return spec.workAreaParameters.lastRealArea, spec.workAreaParameters.lastArea

end



function CutterMultiFruit:onStartWorkAreaProcessing(dt)

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

	local combineVehicle = self:getCombine()

	spec.workAreaParameters.combineVehicle = combineVehicle
	spec.workAreaParameters.lastRealArea = 0
	spec.workAreaParameters.lastThreshedArea = 0
	spec.workAreaParameters.lastStatsArea = 0
	spec.workAreaParameters.lastArea = 0
	spec.workAreaParameters.lastGrowthState = 0
	spec.workAreaParameters.lastGrowthStateArea = 0

	if spec.workAreaParameters.lastFruitType == nil then

		spec.workAreaParameters.fruitTypesToUse = spec.fruitTypes

	else

		spec.workAreaParameters.fruitTypesToUse = {}
		spec.workAreaParameters.fruitTypesToUse[spec.workAreaParameters.lastFruitType] = true

	end

	spec.workAreaParameters.lastFruitType = nil
	spec.isWorking = false

end



function CutterMultiFruit:onEndWorkAreaProcessing(dt, hasProcessed)

	if self.isServer then

		local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

		local lastRealArea = spec.workAreaParameters.lastRealArea
		local lastThreshedArea = spec.workAreaParameters.lastThreshedArea
		local lastStatsArea = spec.workAreaParameters.lastStatsArea
		local lastArea = spec.workAreaParameters.lastArea

		if lastRealArea > 0 then

			if spec.workAreaParameters.combineVehicle ~= nil then

				local inputFruitType = spec.workAreaParameters.lastFruitType

				-- always use the same input fruit type while the ai is active to prevent situations where the combine is unable to unload a different fruit type
				if self:getIsAIActive() then

					local requirements = self:getAIFruitRequirements()
					local requirement = requirements[1]

					if #requirements == 1 and  requirement ~= nil and requirement.fruitType ~= FruitType.UNKNOWN then

						inputFruitType = requirement.fruitType

					end

				end

				-- take fill type conversion into consideration
				local conversionFactor = 1.0
				local outputFillType = g_fruitTypeManager:getFillTypeIndexByFruitTypeIndex(inputFruitType)

				if spec.fruitTypeConverters[inputFruitType] ~= nil then

					outputFillType = spec.fruitTypeConverters[inputFruitType].fillTypeIndex
					conversionFactor = spec.fruitTypeConverters[inputFruitType].conversionFactor

				end

				if spec.lastOutputFillTypes[outputFillType] == nil then

					spec.lastOutputFillTypes[outputFillType] = lastRealArea

				else

					spec.lastOutputFillTypes[outputFillType] = spec.lastOutputFillTypes[outputFillType] + lastRealArea

				end

				if spec.lastPrioritizedOutputType ~= FillType.UNKNOWN then

					outputFillType = spec.lastPrioritizedOutputType

				end

				lastRealArea = lastRealArea * conversionFactor

				local farmId = self:getLastTouchedFarmlandFarmId()
				local appliedDelta = spec.workAreaParameters.combineVehicle:addCutterArea(lastArea, lastRealArea, inputFruitType, outputFillType, spec.strawRatio, farmId)

				if appliedDelta > 0 then

					spec.lastValidInputFruitType = inputFruitType

				end

			end

			local ha = MathUtil.areaToHa(lastStatsArea, g_currentMission:getFruitPixelsToSqm()) -- 4096px are mapped to 2048m
			local stats = g_currentMission:farmStats(self:getLastTouchedFarmlandFarmId())

			stats:updateStats("threshedHectares", ha)
			stats:updateStats("workedHectares", ha)
			spec.lastAreaBiggerZero = lastArea > 0

			if spec.lastAreaBiggerZero then

				spec.lastAreaBiggerZeroTime = g_currentMission.time

			end

			if spec.lastAreaBiggerZero ~= spec.lastAreaBiggerZeroSent then

				self:raiseDirtyFlags(spec.dirtyFlag)
				spec.lastAreaBiggerZeroSent = spec.lastAreaBiggerZero

			end

			if spec.currentInputFruitType ~= spec.currentInputFruitTypeSent then

				self:raiseDirtyFlags(spec.effectDirtyFlag)
				spec.currentInputFruitTypeSent = spec.currentInputFruitType

			end

			if self:getAllowCutterAIFruitRequirements() then

				if self.setAIFruitRequirements ~= nil then

					-- we do not allow changes of the required type while working, just on ai start (if the cutter has more than one requirement or no requirement -> only one requirement allowed at the time)
					-- prevents fruit type changes on field borders
					local requirements = self:getAIFruitRequirements()
					local requirement = requirements[1]

					if #requirements > 1 or requirement == nil or requirement.fruitType == FruitType.UNKNOWN then

						local fruitType = g_fruitTypeManager:getFruitTypeByIndex(spec.currentInputFruitTypeAI)

						if fruitType ~= nil then

							local minState = spec.allowsForageGrowthState and fruitType.minForageGrowthState or fruitType.minHarvestingGrowthState

							self:setAIFruitRequirements(spec.currentInputFruitTypeAI, minState, fruitType.maxHarvestingGrowthState)

						end

					end

				end

				spec.aiNoValidGroundTimer = 0

			end

		else

			if self:getAllowCutterAIFruitRequirements() then

				if hasProcessed then

					if self:getIsAIActive() and self:getRootVehicle():getAILastAllowedToDrive() then

						spec.aiNoValidGroundTimer = spec.aiNoValidGroundTimer + dt

						if spec.aiNoValidGroundTimer > 5000 then

							local rootVehicle = self:getRootVehicle()

							if rootVehicle.stopAIVehicle ~= nil then

								rootVehicle:stopAIVehicle(AIVehicle.STOP_REASON_UNKOWN)

							end

						end

					else

						spec.aiNoValidGroundTimer = 0

					end

				end

			end

		end

	end

end



function CutterMultiFruit:loadTestAreaFromXML(testArea, xmlFile, key)

	XMLUtil.checkDeprecatedXMLElements(xmlFile, self.configFileName, key .. "#startIndex", key .. "#startNode") --FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(xmlFile, self.configFileName, key .. "#widthIndex", key .. "#widthNode") --FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(xmlFile, self.configFileName, key .. "#heightIndex", key .. "#heightNode") --FS17 to FS19

	local start = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. "#startNode"), self.i3dMappings)
	local width = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. "#widthNode"), self.i3dMappings)
	local height = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. "#heightNode"), self.i3dMappings)

	if start ~= nil and width ~= nil and height ~= nil then

		testArea.start = start
		testArea.width = width
		testArea.height = height
		testArea.hasFruitContact = false
		testArea.hasFruitContactSent = false

		return true

	end

	return false

end



function CutterMultiFruit:getIsTestAreaActive(area)

	return true

end



function CutterMultiFruit:getCutterLoad()

	return self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"].cutterLoad

end



function CutterMultiFruit:loadSpeedRotatingPartFromXML(superFunc, speedRotatingPart, xmlFile, key)

	if not superFunc(self, speedRotatingPart, xmlFile, key) then

		return false

	end

	speedRotatingPart.rotateIfTurnedOn = Utils.getNoNil(getXMLBool(xmlFile, key .. "#rotateIfTurnedOn"), false)

	return true

end



function CutterMultiFruit:getIsSpeedRotatingPartActive(superFunc, speedRotatingPart)

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

	if (not spec.allowCuttingWhileRaised and not self:getIsLowered(true)) or (speedRotatingPart.rotateIfTurnedOn and not self:getIsTurnedOn()) then

		return false

	end

	return superFunc(self, speedRotatingPart)

end



function CutterMultiFruit:loadRandomlyMovingPartFromXML(superFunc, part, xmlFile, key)

	local retValue = superFunc(self, part, xmlFile, key)

	part.moveOnlyIfCutted = Utils.getNoNil(getXMLBool(xmlFile, key .. "#moveOnlyIfCutted"), false)

	return retValue

end



function CutterMultiFruit:getIsRandomlyMovingPartActive(superFunc, part)

	local retValue = superFunc(self, part)

	if part.moveOnlyIfCutted then

		retValue = retValue and (self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"].lastAreaBiggerZeroTime >= (g_currentMission.time - 150))

	end

	return retValue

end



function CutterMultiFruit:getIsWorkAreaActive(superFunc, workArea)

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

	if self.getAllowsLowering == nil or self:getAllowsLowering() then

		if not spec.allowCuttingWhileRaised and not self:getIsLowered(true) then

			return false

		end

	end

	return superFunc(self, workArea)

end



function CutterMultiFruit:doCheckSpeedLimit(superFunc)

	return superFunc(self) or (self:getIsTurnedOn() and (self.getIsLowered == nil or self:getIsLowered()))

end



function CutterMultiFruit:loadWorkAreaFromXML(superFunc, workArea, xmlFile, key)

	local retValue = superFunc(self, workArea, xmlFile, key)

	workArea.chopperAreaIndex = getXMLInt(xmlFile, key..".chopperArea#index")

	return retValue

end



function CutterMultiFruit:getDirtMultiplier(superFunc)

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

	if spec.isWorking then

		return superFunc(self) + self:getWorkDirtMultiplier() * self:getLastSpeed() / self.speedLimit

	end

	return superFunc(self)

end



function CutterMultiFruit:getWearMultiplier(superFunc)

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

	if spec.isWorking then

		return superFunc(self) + self:getWorkWearMultiplier() * self:getLastSpeed() / self.speedLimit

	end

	return superFunc(self)

end



function CutterMultiFruit:onPreAttach(attacherVehicle, inputJointDescIndex, jointDescIndex)

	if self.isClient then

		CutterMultiFruit.updateExtraObjects(self)

	end

end



function CutterMultiFruit:onPostDetach(attacherVehicle, implement)

	if self.isClient then

		CutterMultiFruit.updateExtraObjects(self)

	end

end



function CutterMultiFruit:onTurnedOn()

	if self.isClient then

		local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

		g_animationManager:startAnimations(spec.animationNodes)

	end

end



function CutterMultiFruit:onTurnedOff()

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

	if self.isClient then

		g_animationManager:stopAnimations(spec.animationNodes)
		g_effectManager:resetEffects(spec.currentCutterEffect)

		spec.currentInputFruitType = FruitType.UNKNOWN
		spec.currentInputFruitTypeAI = FruitType.UNKNOWN
		spec.currentInputFillType = FillType.UNKNOWN

	end

end



function CutterMultiFruit:onAIImplementStart()

	-- clear ai fruit requirements on start of ai vehicle to get the newest fruit type in front of the cutter
	if self:getAllowCutterAIFruitRequirements() then

		self:clearAIFruitRequirements()

		-- default require all fruit types the cutter can handle so we don't get other fruit types
		local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

		for fruitTypeIndex, _ in pairs(spec.fruitTypes) do

			local fruitType = g_fruitTypeManager:getFruitTypeByIndex(fruitTypeIndex)

			if fruitType ~= nil then

				local minState = spec.allowsForageGrowthState and fruitType.minForageGrowthState or fruitType.minHarvestingGrowthState

				self:addAIFruitRequirement(fruitType.index, minState, fruitType.maxHarvestingGrowthState)

			end

		end

	end

end



function CutterMultiFruit.updateExtraObjects(self)

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

	if spec.currentInputFruitType ~= nil then

		local extraObject = spec.fruitExtraObjects[spec.currentInputFruitType]

		if spec.hideExtraObjectsOnDetach then

			if self.getAttacherVehicle == nil or self:getAttacherVehicle() == nil then

				extraObject = nil

			end

		end

		if extraObject ~= spec.currentExtraObject then

			if spec.currentExtraObject ~= nil then

				if spec.currentExtraObject.node ~= nil then

					setVisibility(spec.currentExtraObject.node, false)

				end

				if spec.currentExtraObject.anim ~= nil and self.playAnimation ~= nil then

					self:playAnimation(spec.currentExtraObject.anim, -1, self:getAnimationTime(spec.currentExtraObject.anim), true)

				end

				spec.currentExtraObject = nil

			end

			if extraObject ~= nil then

				if extraObject.node ~= nil then

					setVisibility(extraObject.node, true)

				end

				if extraObject.anim ~= nil and self.playAnimation ~= nil then

					self:playAnimation(extraObject.anim, 1, self:getAnimationTime(extraObject.anim), true)

				end

				spec.currentExtraObject = extraObject

			end

		end

	end

end



function CutterMultiFruit.getDefaultSpeedLimit()

	return 10

end



function CutterMultiFruit:updateDebugValues(values)

	local spec = self["spec_" .. CutterMultiFruit.modName .. ".CutterMultiFruit"]

	table.insert(values, {name="lastPrioritizedOutputType", value=string.format("%s", g_fillTypeManager:getFillTypeNameByIndex(spec.lastPrioritizedOutputType))})

	local sum = 0

	for fillType, value in pairs(spec.lastOutputFillTypes) do

		sum = sum + value

	end

	for fillType, value in pairs(spec.lastOutputFillTypes) do

		table.insert(values, {name=string.format("buffer (%s)", g_fillTypeManager:getFillTypeNameByIndex(fillType)), value=string.format("%.0f%%", value/sum*100)})

	end

end