
OverwriteFillUnitMultiFruit = {}
OverwriteFillUnitMultiFruit.modDir = g_currentModDirectory
OverwriteFillUnitMultiFruit.modName = g_currentModName

local modDesc = loadXMLFile("modDesc", OverwriteFillUnitMultiFruit.modDir .. "modDesc.xml")

OverwriteFillUnitMultiFruit.usedXmlString = Utils.getNoNil(getXMLString(modDesc, "modDesc.LuaVariable#name"), "")

addModEventListener(OverwriteFillUnitMultiFruit)



function OverwriteFillUnitMultiFruit:loadMap(name)

	local fruitTypesString = ""

	for fruitTypeIndex, fruitTypeData in pairs(g_fruitTypeManager.fruitTypes) do

		if fruitTypeData ~= nil then

			local minHarvestingGrowthState = fruitTypeData.minHarvestingGrowthState
			local minPreparingGrowthState = fruitTypeData.minPreparingGrowthState
			local preparedGrowthState = fruitTypeData.preparedGrowthState
			local name = fruitTypeData.name

			if minHarvestingGrowthState == 9 and minPreparingGrowthState ~= -1 and preparedGrowthState == 9 and name ~= "SUGARCANE" then

				if fruitTypesString ~= "" then 
					fruitTypesString = fruitTypesString .." "
				end

				fruitTypesString = fruitTypesString..tostring(name)

			end

		end

	end

	for _, item in pairs(g_storeManager.items) do

		if item.customEnvironment == OverwriteFillUnitMultiFruit.modName then

			if item.specs.fillTypes.fillTypeNames == OverwriteFillUnitMultiFruit.usedXmlString then

				item.specs.fillTypes.fillTypeNames = fruitTypesString

			end

			if item.specs.fillTypes.fruitTypeNames == OverwriteFillUnitMultiFruit.usedXmlString then

				item.specs.fillTypes.fruitTypeNames = fruitTypesString

			end

		end

	end

end



function OverwriteFillUnitMultiFruit:deleteMap()
end

function OverwriteFillUnitMultiFruit:keyEvent(unicode, sym, modifier, isDown)
end

function OverwriteFillUnitMultiFruit:mouseEvent(posX, posY, isDown, isUp, button)
end

function OverwriteFillUnitMultiFruit:update(dt)
end

function OverwriteFillUnitMultiFruit:draw()
end



-- fillUnit overwritten Function

function OverwriteFillUnitMultiFruit.overwrite_fillUnit_loadFillUnitFromXML(self, superFunc, xmlFile, key, entry, index)

	local spec = self.spec_fillUnit

	entry.fillUnitIndex = index
	entry.capacity = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#capacity"), math.huge)
	entry.updateMass = Utils.getNoNil(getXMLBool(xmlFile, key .. "#updateMass"), true)
	entry.canBeUnloaded = Utils.getNoNil(getXMLBool(xmlFile, key .. "#canBeUnloaded"), true)
	entry.needsSaving = true
	entry.fillLevel = 0
	entry.fillLevelSent = 0
	entry.fillType = FillType.UNKNOWN
	entry.fillTypeSent = FillType.UNKNOWN
	entry.fillTypeToDisplay = FillType.UNKNOWN
	entry.fillLevelToDisplay = nil
	entry.lastValidFillType = FillType.UNKNOWN
	entry.lastValidFillTypeSent = FillType.UNKNOWN



	if hasXMLProperty(xmlFile, key .. ".exactFillRootNode") then

		XMLUtil.checkDeprecatedXMLElements(xmlFile, self.configFileName, key .. ".exactFillRootNode#index", key .. ".exactFillRootNode#node") --FS17 to FS19
		entry.exactFillRootNode = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. ".exactFillRootNode#node"), self.i3dMappings)

		if entry.exactFillRootNode ~= nil then

			local colMask = getCollisionMask(entry.exactFillRootNode)

			if bitAND(FillUnit.EXACTFILLROOTNODE_MASK, colMask) == 0 then

				g_logManager:xmlWarning(self.configFileName, "Invalid collision mask for exactFillRootNode '%s'. Bit 30 needs to be set!", key)
				return false

			end

			spec.exactFillRootNodeToFillUnit[entry.exactFillRootNode] = entry
			spec.exactFillRootNodeToExtraDistance[entry.exactFillRootNode] = getXMLFloat(xmlFile, key .. ".exactFillRootNode#extraEffectDistance") or 0
			spec.hasExactFillRootNodes = true

		else

			g_logManager:xmlWarning(self.configFileName, "ExactFillRootNode not found for fillUnit '%s'!", key)

		end

	end



	XMLUtil.checkDeprecatedXMLElements(xmlFile, self.configFileName, key .. ".autoAimTargetNode#index", key .. ".autoAimTargetNode#node") --FS17 to FS19
	entry.autoAimTarget = {}
	entry.autoAimTarget.node = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. ".autoAimTargetNode#node"), self.i3dMappings)

	if entry.autoAimTarget.node ~= nil then

		entry.autoAimTarget.baseTrans = {getTranslation(entry.autoAimTarget.node)}
		entry.autoAimTarget.startZ = getXMLFloat(xmlFile,  key ..".autoAimTargetNode#startZ")
		entry.autoAimTarget.endZ = getXMLFloat(xmlFile,  key ..".autoAimTargetNode#endZ")
		entry.autoAimTarget.startPercentage = Utils.getNoNil(getXMLFloat(xmlFile,  key ..".autoAimTargetNode#startPercentage"), 25)/100
		entry.autoAimTarget.invert = Utils.getNoNil(getXMLBool(xmlFile,  key ..".autoAimTargetNode#invert"), false)

		if entry.autoAimTarget.startZ ~= nil and entry.autoAimTarget.endZ ~= nil then

			local startZ = entry.autoAimTarget.startZ

			if entry.autoAimTarget.invert then

				startZ = entry.autoAimTarget.endZ

			end

			setTranslation(entry.autoAimTarget.node, entry.autoAimTarget.baseTrans[1], entry.autoAimTarget.baseTrans[2], startZ)

		end

	end



	entry.supportedFillTypes = {}

	local fillTypes
	local fillTypeCategories = getXMLString(xmlFile, key .. "#fillTypeCategories")
	local fillTypeNames = getXMLString(xmlFile, key .. "#fillTypes")

	if fillTypeNames == FruitPreparerMultiFruit.usedXmlString then

		fillTypeNames = ""

		for fruitTypeIndex, fruitTypeData in pairs(g_fruitTypeManager.fruitTypes) do

			if fruitTypeData ~= nil then

				local minHarvestingGrowthState = fruitTypeData.minHarvestingGrowthState
				local minPreparingGrowthState = fruitTypeData.minPreparingGrowthState
				local preparedGrowthState = fruitTypeData.preparedGrowthState
				local name = fruitTypeData.name

				if minHarvestingGrowthState == 9 and minPreparingGrowthState ~= -1 and preparedGrowthState == 9 and name ~= "SUGARCANE" then

					if fillTypeNames ~= "" then 
						fillTypeNames = fillTypeNames .." "
					end

					fillTypeNames = fillTypeNames..tostring(name)

				end

			end

		end

	end

	if fillTypeCategories ~= nil and fillTypeNames == nil then

		fillTypes = g_fillTypeManager:getFillTypesByCategoryNames(fillTypeCategories, "Warning: '"..self.configFileName.. "' has invalid fillTypeCategory '%s'.")

	elseif fillTypeCategories == nil and fillTypeNames ~= nil then

		fillTypes = g_fillTypeManager:getFillTypesByNames(fillTypeNames, "Warning: '"..self.configFileName.. "' has invalid fillType '%s'.")

	else

		g_logManager:xmlWarning(self.configFileName, "Missing 'fillTypeCategories' or 'fillTypes' for fillUnit '%s'", key)
		return false

	end

	if fillTypes ~= nil then

		for _,fillType in pairs(fillTypes) do

			entry.supportedFillTypes[fillType] = true

		end

	end



	entry.supportedToolTypes = {}

	for i=1, g_toolTypeManager:getNumberOfToolTypes() do

		entry.supportedToolTypes[i] = true

	end

	local startFillLevel = getXMLFloat(xmlFile, key .. "#startFillLevel")
	local startFillTypeStr = getXMLString(xmlFile, key .. "#startFillType")

	if startFillTypeStr ~= nil then

		local startFillTypeIndex = g_fillTypeManager:getFillTypeIndexByName(startFillTypeStr)

		if startFillTypeIndex ~= nil then

			entry.startFillLevel = startFillLevel
			entry.startFillTypeIndex = startFillTypeIndex

		end

	end

	entry.fillRootNode = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. ".fillRootNode#node"), self.i3dMappings)

	if entry.fillRootNode == nil then

		entry.fillRootNode = self.components[1].node

	end

	entry.fillMassNode = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. ".fillMassNode#node"), self.i3dMappings)

	local updateFillLevelMass = Utils.getNoNil(getXMLBool(xmlFile, key .. "#updateFillLevelMass"), true)

	if entry.fillMassNode == nil and updateFillLevelMass then

		entry.fillMassNode = self.components[1].node

	end



	-- mp sync info
	entry.synchronizeFillLevel = Utils.getNoNil(getXMLBool(xmlFile, key .. "#synchronizeFillLevel"), true)
	entry.synchronizeFullFillLevel = Utils.getNoNil(getXMLBool(xmlFile, key .. "#synchronizeFullFillLevel"), false)

	local defaultBits = 16

	for startCapacity, bits in pairs(FillUnit.CAPACITY_TO_NETWORK_BITS) do

		if entry.capacity >= startCapacity then

			defaultBits = bits

		end

	end

	entry.synchronizationNumBits = Utils.getNoNil(getXMLInt(xmlFile, key .. "#synchronizationNumBits"), defaultBits)
	entry.showOnHud = Utils.getNoNil(getXMLBool(xmlFile, key .. "#showOnHud"), true)
	entry.blocksAutomatedTrainTravel = Utils.getNoNil(getXMLBool(xmlFile, key .. "#blocksAutomatedTrainTravel"), false)
	entry.fillAnimation = getXMLString(xmlFile, key .. "#fillAnimation")
	entry.fillAnimationLoadTime = getXMLFloat(xmlFile, key .. "#fillAnimationLoadTime")
	entry.fillAnimationEmptyTime = getXMLFloat(xmlFile, key .. "#fillAnimationEmptyTime")



	if self.isClient then

		entry.alarmTriggers = {}

		local i = 0

		while true do

			local nodeKey = key .. string.format(".alarmTriggers.alarmTrigger(%d)", i)

			if not hasXMLProperty(xmlFile, nodeKey) then
				break
			end

			local alarmTrigger = {}

			if self:loadAlarmTrigger(xmlFile, nodeKey, alarmTrigger, entry) then

				table.insert(entry.alarmTriggers, alarmTrigger)

			end

			i = i + 1

		end

		entry.measurementNodes = {}

		i = 0

		while true do

			local nodeKey = key .. string.format(".measurementNodes.measurementNode(%d)", i)

			if not hasXMLProperty(xmlFile, nodeKey) then
				break
			end

			local measurementNode = {}

			if self:loadMeasurementNode(xmlFile, nodeKey, measurementNode) then

				table.insert(entry.measurementNodes, measurementNode)

			end

			i = i + 1

		end

		entry.fillPlane = {}
		entry.lastFillPlaneType = nil

		if not self:loadFillPlane(xmlFile, key..".fillPlane", entry.fillPlane, entry) then

			entry.fillPlane = nil

		end

		entry.fillEffects = g_effectManager:loadEffect(xmlFile, key .. ".fillEffect", self.components, self, self.i3dMappings)
		entry.animationNodes = g_animationManager:loadAnimations(xmlFile, key..".animationNodes", self.components, self, self.i3dMappings)
		XMLUtil.checkDeprecatedXMLElements(xmlFile, self.configFileName, key .. ".fillLevelHud", key .. ".dashboard") --FS17 to FS19

		if self.loadDashboardsFromXML ~= nil then

			self:loadDashboardsFromXML(xmlFile, key, {valueTypeToLoad = "fillLevel",
													  valueObject = entry,
													  valueFunc = "fillLevel",
													  minFunc = 0,
													  maxFunc = "capacity"})

			self:loadDashboardsFromXML(xmlFile, key, {valueTypeToLoad = "fillLevelPct",
													  valueObject = entry,
													  valueFunc = function(fillUnit)
														  return MathUtil.clamp(fillUnit.fillLevel/fillUnit.capacity, 0, 1) * 100
													  end,
													  minFunc = 0,
													  maxFunc = 100})

			self:loadDashboardsFromXML(xmlFile, key, {valueTypeToLoad = "fillLevelWarning",
													  valueObject = entry,
													  valueFunc = "fillLevel",
													  minFunc = 0,
													  maxFunc = "capacity",
													  additionalAttributesFunc = Dashboard.warningAttributes,
													  stateFunc = Dashboard.warningState})

		end

	end

	-- for _, item in pairs(g_storeManager.items) do

		-- if item.customEnvironment == OverwriteFillUnitMultiFruit.modName then

			-- if item.specs.fillTypes.fillTypeNames == OverwriteFillUnitMultiFruit.usedXmlString then

				-- item.specs.fillTypes.fillTypeNames = FruitPreparerMultiFruit.fruitTypesString

			-- end

			-- if item.specs.fillTypes.fruitTypeNames == OverwriteFillUnitMultiFruit.usedXmlString then

				-- item.specs.fillTypes.fruitTypeNames = FruitPreparerMultiFruit.fruitTypesString

			-- end

		-- end

	-- end

	return true

end

FillUnit.loadFillUnitFromXML = Utils.overwrittenFunction(FillUnit.loadFillUnitFromXML, OverwriteFillUnitMultiFruit.overwrite_fillUnit_loadFillUnitFromXML )